'''Registry to store all the cast functions to cast an object from type A to type B.'''


from functools import wraps


__all__ = ['cast', 'castable', 'register_cast', 'register_castable']


# map: (src_type, dst_type) -> cast function
_cast_registry = {}

_castable_registry = {}
# map: (src_type, dst_type) -> castable function


def castable(obj, dst_type):
    '''Checks if an object can be cast to a target type.

    Parameters
    ----------
    obj : object
        object to be cast
    dst_type : type
        type or class to cast to

    Returns
    -------
    bool
        whether or not the object is castable to `dst_type`

    Raises
    ------
    NotImplementedError
        if the castable function has not been registered using decorators `register_castable` or `register_cast`
    '''
    key = (type(obj), dst_type)
    if key in _castable_registry:
        return _castable_registry[key](obj)
    if key in _cast_registry:
        return True
    raise NotImplementedError("Castable function to cast from type {} to type {} not found.".format(key[0], key[1]))


def cast(obj, dst_type):
    '''Casts an object to a target type.

    Parameters
    ----------
    obj : object
        object to be cast
    dst_type : type
        type or class to cast to

    Returns
    -------
    object
        the cast version of `obj`

    Raises
    ------
    NotImplementedError
        if the cast function has not been registered using decorator `register_cast`
    '''
    key = (type(obj), dst_type)
    if key in _cast_registry:
        return _cast_registry[key](obj)
    raise NotImplementedError("Cast function from type {} to type {} not found.".format(key[0], key[1]))


def register_cast(src_type, dst_type):
    '''A decorator to register a function to cast from type A to type B.

    Parameters
    ----------
    src_type : type
        type or class to cast from
    dst_type : type
        type or class to cast to
    '''

    def register_cast_decorator(func):
        @wraps(func)
        def function_wrapper(func):
            key = (src_type, dst_type)
            if key in _cast_registry:
                raise SyntaxError("Cast function from type {} to type {} has been registered before.".format(src_type, dst_type))
            _cast_registry[key] = func

        return function_wrapper

    return register_cast_decorator


def register_castable(src_type, dst_type):
    '''A decorator to register a function to check if we can cast from type A to type B.

    Parameters
    ----------
    src_type : type
        type or class to cast from
    dst_type : type
        type or class to cast to
    '''

    def register_cast_decorator(func):
        @wraps(func)
        def function_wrapper(func):
            key = (src_type, dst_type)
            if key in _castable_registry:
                raise SyntaxError("Castable function from type {} to type {} has been registered before.".format(src_type, dst_type))
            _castable_registry[key] = func

        return function_wrapper

    return register_cast_decorator
